import aerosandbox.numpy as np
import aerosandbox.tools.units as u
from typing import Dict


def modified_DAPCA_IV_production_cost_analysis(
        design_empty_weight: float,
        design_maximum_airspeed: float,
        n_airplanes_produced: int,
        n_engines_per_aircraft: int,
        cost_per_engine: float,
        cost_avionics_per_airplane: float,
        n_pax: int,
        cpi_relative_to_2012_dollars: float = 1.275,  # updated for 2022
        n_flight_test_aircraft: int = 4,
        is_cargo_airplane: bool = False,
        primary_structure_material: str = "aluminum",
        per_passenger_cost_model: str = "general_aviation",
        engineering_wrap_rate_2012_dollars: float = 115.,
        tooling_wrap_rate_2012_dollars: float = 118.,
        quality_control_wrap_rate_2012_dollars: float = 108.,
        manufacturing_wrap_rate_2012_dollars: float = 98.,
) -> Dict[str, float]:
    """
    Computes the cost of an aircraft in present-day dollars, using the Modified DAPCA IV cost model.

    Be sure to adjust `cpi_relative_to_2012_dollars` to the current values in order to accurately model inflation.

    The DAPCA IV cost model is a statistical regression of historical aircraft cost data. It provides reasonable
    results for most classes of aircraft, including transports, fighters, bombers, and even GA and UAV aircraft with
    suitable adjustments.

    It was created by the RAND Corporation.

    The Modified DAPCA IV cost model is a modification of the DAPCA IV cost model that includes additional cost
    estimates for engine cost (as the original DAPCA model assumes that this is known).

    See Raymer, Aircraft Design: A Conceptual Approach, 5th Edition, Section 18.4.2 pg. 711 for more information.

    Args:
        design_empty_weight: The design empty weight of the entire aircraft, in kg.

        design_maximum_airspeed: The design maximum airspeed of the aircraft, in m/s.

        n_airplanes_produced: The number of airplanes produced or the number to be produced in 5 years, whichever is
        less.

        n_engines_per_aircraft: The number of engines per aircraft.

        cost_per_engine: The cost of each engine, in present-day dollars.

        cost_avionics_per_airplane: The cost of avionics per airplane, in present-day dollars.

        n_pax: The number of passengers.

        cpi_relative_to_2012_dollars: The consumer price index at the present day divided by the consumer price index
        in 2012, seasonally-adjusted.

            To quickly find this, use data from the St. Louis Federal Reserve. Below is the CPI, normalized to 2012.
            https://fred.stlouisfed.org/graph/?g=10PU0

            For example, in 2022, one would use 1.275.

        n_flight_test_aircraft: The number of flight test aircraft. Typically 2 to 6.

        is_cargo_airplane: Whether the airplane is a cargo airplane. If so, the quality control cost is lower.

        primary_structure_material: The primary structure material. Options are:
            - "aluminum"
            - "carbon_fiber"
            - "fiberglass"
            - "steel"
            - "titanium"

        per_passenger_cost_model: The per-passenger cost model. Options are:
            - "general_aviation": General aviation aircraft, such as Cessna 172s.
            - "jet_transport": Jet transport aircraft, such as Boeing 737s.
            - "regional_transport": Regional transport aircraft, such as Embraer E175s.

        engineering_wrap_rate_2012_dollars: The engineering wrap rate in 2012 dollars.

        tooling_wrap_rate_2012_dollars: The tooling wrap rate in 2012 dollars.

        quality_control_wrap_rate_2012_dollars: The quality control wrap rate in 2012 dollars.

        manufacturing_wrap_rate_2012_dollars: The manufacturing wrap rate in 2012 dollars.

    Returns:
        A dictionary of costs, to produce all `n_airplanes_produced` airplanes, in present-day dollars.

        Keys and values are as follows:

            - "engineering_labor": Engineering labor cost.

            - "tooling_labor": Tooling labor cost.

            - "manufacturing_labor": Manufacturing labor cost.

            - "quality_control_labor": Quality control labor cost.

            - "development_support": Development support cost. From Raymer: "Includes fabrication of mockups, iron-bird subsystem
            simulators, structural test articles, and other test articles."

            - "flight_test": Flight test cost. From Raymer: "Includes all costs incurred to demonstrate airworthiness
            for civil certification or Mil-Spec compliance except for the costs of the flight-test aircraft
            themselves. Costs for the flight-test aircraft are included in the total production-run cost estimation.
            Includes planning, instrumentation, flight operations, data reduction, and engineering and manufacturing
            support of flight testing."

            - "manufacturing_materials": Manufacturing materials cost. From Raymer: "Includes all raw materials and
            purchased hardware and equipment."

            - "engine_cost": Engine cost.

            - "avionics_cost": Avionics cost.

            - "total_cost": Total cost.

    """
    # Abbreviated constants for readability
    W = design_empty_weight  # kg
    V = design_maximum_airspeed / u.kph  # km/hour
    Q = n_airplanes_produced

    ### Estimate labor hours
    hours = {}

    hours["engineering"] = 5.18 * W ** 0.777 * V ** 0.894 * Q ** 0.163
    hours["tooling"] = 7.22 * W ** 0.777 * V ** 0.696 * Q ** 0.263
    hours["manufacturing"] = 10.5 * W ** 0.82 * V ** 0.484 * Q ** 0.641
    hours["quality_control"] = hours["manufacturing"] * (0.076 if is_cargo_airplane else 0.133)

    ### Account for materials difficulties
    if primary_structure_material == "aluminum":
        materials_hourly_multiplier = 1.0
    elif primary_structure_material == "carbon_fiber":
        materials_hourly_multiplier = (1.1 + 1.8) / 2
    elif primary_structure_material == "fiberglass":
        materials_hourly_multiplier = (1.1 + 1.2) / 2
    elif primary_structure_material == "steel":
        materials_hourly_multiplier = (1.5 + 2.0) / 2
    elif primary_structure_material == "titanium":
        materials_hourly_multiplier = (1.1 + 1.8) / 2
    else:
        raise ValueError("Invalid value of `primary_structure_material`.")

    hours = {
        k: v * materials_hourly_multiplier
        for k, v in hours.items()
    }

    ### Convert labor hours to labor costs in 2012 dollars
    costs_2012_dollars = {}

    costs_2012_dollars["engineering_labor"] = hours["engineering"] * engineering_wrap_rate_2012_dollars
    costs_2012_dollars["tooling_labor"] = hours["tooling"] * tooling_wrap_rate_2012_dollars
    costs_2012_dollars["manufacturing_labor"] = hours["manufacturing"] * manufacturing_wrap_rate_2012_dollars
    costs_2012_dollars["quality_control_labor"] = hours["quality_control"] * quality_control_wrap_rate_2012_dollars

    costs_2012_dollars["development_support"] = 67.4 * W ** 0.630 * V ** 1.3
    costs_2012_dollars["flight_test"] = 1947 * W ** 0.325 * V ** 0.822 * n_flight_test_aircraft ** 1.21
    costs_2012_dollars["manufacturing_materials"] = 31.2 * W ** 0.921 * V ** 0.621 * Q ** 0.799

    ### Add in the per-passenger cost for aircraft interiors:
    # Seats, luggage bins, closets, lavatories, insulation, ceilings, floors, walls, etc.
    # Costs are from Raymer, Aircraft Design: A Conceptual Approach, 5th edition. Section 18.4.2, page 715.
    if per_passenger_cost_model == "general_aviation":
        costs_2012_dollars["aircraft_interiors"] = n_airplanes_produced * n_pax * 850
    elif per_passenger_cost_model == "jet_transport":
        costs_2012_dollars["aircraft_interiors"] = n_airplanes_produced * n_pax * 3500
    elif per_passenger_cost_model == "regional_transport":
        costs_2012_dollars["aircraft_interiors"] = n_airplanes_produced * n_pax * 1700
    else:
        raise ValueError(f"Invalid value of `per_passenger_cost_model`!")

    ### Convert all costs to present-day dollars
    costs = {
        k: v * cpi_relative_to_2012_dollars
        for k, v in costs_2012_dollars.items()
    }

    ### Add engine and avionics costs
    costs["engine_cost"] = cost_per_engine * n_engines_per_aircraft * n_airplanes_produced
    costs["avionics"] = cost_avionics_per_airplane * n_airplanes_produced

    ### Total all costs and return
    costs["total"] = sum(costs.values())

    return costs



from aerosandbox.common import ExplicitAnalysis
import aerosandbox.numpy as np
import subprocess
from pathlib import Path
from aerosandbox.geometry import Airfoil
from typing import Union, List, Dict
import tempfile
import warnings
import os


class XFoil(ExplicitAnalysis):
    """

    An interface to XFoil, a 2D airfoil analysis tool developed by Mark Drela at MIT.

    Requires XFoil to be on your computer; XFoil is available here: https://web.mit.edu/drela/Public/web/xfoil/

    It is recommended (but not required) that you add XFoil to your system PATH environment variable such that it can
    be called with the command `xfoil`. If this is not the case, you need to specify the path to your XFoil
    executable using the `xfoil_command` argument of the constructor.

    Usage example:

    >>> xf = XFoil(
    >>>     airfoil=Airfoil("naca2412").repanel(n_points_per_side=100),
    >>>     Re=1e6,
    >>> )
    >>>
    >>> result_at_single_alpha = xf.alpha(5)
    >>> result_at_several_CLs = xf.cl([0.5, 0.7, 0.8, 0.9])
    >>> result_at_multiple_alphas = xf.alpha([3, 5, 60]) # Note: if a result does not converge (such as the 60 degree case here), it will not be included in the results.

    """

    def __init__(self,
                 airfoil: Airfoil,
                 Re: float = 0.,
                 mach: float = 0.,
                 n_crit: float = 9.,
                 xtr_upper: float = 1.,
                 xtr_lower: float = 1.,
                 hinge_point_x: float = None,
                 full_potential: bool = False,
                 max_iter: int = 100,
                 xfoil_command: str = "xfoil",
                 xfoil_repanel: bool = True,
                 verbose: bool = False,
                 timeout: Union[float, int, None] = 30,
                 working_directory: str = None,
                 ):
        """
        Interface to XFoil. Compatible with both XFoil v6.xx (public) and XFoil v7.xx (private, contact Mark Drela at
        MIT for a copy.)

        Args:

            airfoil: The angle of attack [degrees]

            Re: The chord-referenced Reynolds number

            mach: The freestream Mach number

            n_crit: The critical Tollmein-Schlichting wave amplification factor

            xtr_upper: The upper-surface trip location [x/c]

            xtr_lower: The lower-surface trip location [x/c]

            hinge_point_x: The x/c location of the hinge point. This is used to calculate hinge moment. If this is
            None, hinge moment is not calculated.

            full_potential: If this is set True, it will turn full-potential mode on. Note that full-potential mode
            is only available in XFoil v7.xx or higher. (Unless you have specifically gone through the trouble of
            acquiring a copy of XFoil v7.xx you likely have v6.xx. Version 7.xx is not publicly distributed as of
            2022; contact Mark Drela at MIT for a copy.) Note that if you enable this flag with XFoil v6.xx,
            you'll likely get an error (no output file generated).

            max_iter: How many iterations should we let XFoil do?

            xfoil_command: The command-line argument to call XFoil.

                * If XFoil is on your system PATH, then you can just leave this as "xfoil".

                * If XFoil is not on your system PATH, then you should provide a filepath to the XFoil executable.

                Note that XFoil is not on your PATH by default. To tell if XFoil is not on your system PATH,
                open up a terminal and type "xfoil".

                    * If the XFoil menu appears, it's on your PATH.

                    * If you get something like "'xfoil' is not recognized as an internal or external command..." or
                    "Command 'xfoil' not found, did you mean...", then it is not on your PATH and you'll need to
                    specify the location of your XFoil executable as a string.

                To add XFoil to your path, modify your system's environment variables. (Google how to do this for
                your OS.)

            xfoil_repanel: Controls whether to allow XFoil to repanel your airfoil using its internal methods (PANE
            -> PPAR, both with default settings, 160 nodes)

            verbose: Controls whether or not XFoil output is printed to command line.

            timeout: Controls how long any individual XFoil run (i.e. alpha sweep) is allowed to run before the
            process is killed. Given in units of seconds. To disable timeout, set this to None.

            working_directory: Controls which working directory is used for the XFoil input and output files. By
            default, this is set to a TemporaryDirectory that is deleted after the run. However, you can set it to
            somewhere local for debugging purposes.

        """
        if mach >= 1:
            raise ValueError("XFoil will terminate if a supersonic freestream Mach number is given.")

        self.airfoil = airfoil
        self.Re = Re
        self.mach = mach
        self.n_crit = n_crit
        self.xtr_upper = xtr_upper
        self.xtr_lower = xtr_lower
        self.hinge_point_x = hinge_point_x
        self.full_potential = full_potential
        self.max_iter = max_iter
        self.xfoil_command = xfoil_command
        self.xfoil_repanel = xfoil_repanel
        self.verbose = verbose
        self.timeout = timeout
        self.working_directory = working_directory

    def _default_keystrokes(self) -> List[str]:
        """
        Returns a list of XFoil keystrokes that are common to all XFoil runs.

        Returns:
            A list of strings, each of which is a single XFoil keystroke to be followed by <enter>.
        """
        run_file_contents = []

        # Disable graphics
        run_file_contents += [
            "plop",
            "g",
            "",
        ]

        if self.xfoil_repanel:
            run_file_contents += [
                "pane",
                "ppar",
                "",
            ]

        # Enter oper mode
        run_file_contents += [
            "oper",
        ]

        # Handle Re
        if self.Re != 0:
            run_file_contents += [
                f"v {self.Re}",
            ]

        # Handle mach
        if self.mach != 0:
            run_file_contents += [
                f"m {self.mach}",
            ]

        # Handle hinge moment
        if self.hinge_point_x is not None:
            run_file_contents += [
                "hinc",
                f"fnew {self.hinge_point_x} {self.airfoil.local_camber(self.hinge_point_x)}",
                "fmom",
            ]

        if self.full_potential:
            run_file_contents += [
                "full",
                "fpar",
                f"i {self.max_iter}",
                "",
            ]

        # Handle iterations
        run_file_contents += [
            f"iter {self.max_iter}",
        ]

        # Handle trips and ncrit
        if not (self.xtr_upper == 1 and self.xtr_lower == 1 and self.n_crit == 9):
            run_file_contents += [
                "vpar",
                f"xtr {self.xtr_upper} {self.xtr_lower}",
                f"n {self.n_crit}",
                "",
            ]

        # Set polar accumulation
        run_file_contents += [
            "pacc",
            "",
            "",
        ]

        # Include more data in polar
        run_file_contents += [
            "cinc"  # include minimum Cp
        ]

        return run_file_contents

    def _run_xfoil(self,
                   run_command: str,
                   ) -> Dict[str, np.ndarray]:
        """
        Private function to run XFoil.

        Args: run_command: A string with any XFoil keystroke inputs that you'd like. By default, you start off within the OPER
        menu. All of the inputs indicated in the constructor have been set already, but you can override them here (for
        this run only) if you want.

        Returns: A dictionary containing all converged solutions obtained with your inputs.

        """
        # Set up a temporary directory
        with tempfile.TemporaryDirectory() as directory:
            directory = Path(directory)

            ### Alternatively, work in another directory:
            if self.working_directory is not None:
                directory = Path(self.working_directory)  # For debugging

            # Designate an intermediate file for file I/O
            output_filename = "output.txt"

            # Handle the airfoil file
            airfoil_file = "airfoil.dat"
            self.airfoil.write_dat(directory / airfoil_file)

            # Handle the keystroke file
            keystrokes = self._default_keystrokes()
            keystrokes += [run_command]
            keystrokes += [
                "pwrt",
                f"{output_filename}",
                "",
                "",
                "quit"
            ]

            # Remove an old output file, if one exists:
            try:
                os.remove(directory / output_filename)
            except FileNotFoundError:
                pass

            ### Execute
            try:
                command = f'{self.xfoil_command} {airfoil_file}'
                proc = subprocess.Popen(
                    command,
                    cwd=directory,
                    stdin=subprocess.PIPE,
                    stdout=None if self.verbose else subprocess.DEVNULL,
                    stderr=None if self.verbose else subprocess.DEVNULL,
                    text=True,
                    # shell=True,
                    # timeout=self.timeout,
                    # check=True
                )
                outs, errs = proc.communicate(
                    input="\n".join(keystrokes),
                    timeout=self.timeout
                )
                return_code = proc.poll()

            except subprocess.TimeoutExpired:
                proc.kill()
                outs, errs = proc.communicate()

                warnings.warn(
                    "XFoil run timed out!\n"
                    "If this was not expected, try increasing the `timeout` parameter\n"
                    "when you create this AeroSandbox XFoil instance.",
                    stacklevel=2
                )
            except subprocess.CalledProcessError as e:
                if e.returncode == 11:
                    raise RuntimeError(
                        "XFoil segmentation-faulted. This is likely because your input airfoil has too many points.\n"
                        "Try repaneling your airfoil with `Airfoil.repanel()` before passing it into XFoil.\n"
                        "For further debugging, turn on the `verbose` flag when creating this AeroSandbox XFoil instance.")
                elif e.returncode == 8 or e.returncode == 136:
                    raise RuntimeError(
                        "XFoil returned a floating point exception. This is probably because you are trying to start\n"
                        "your analysis at an operating point where the viscous boundary layer can't be initialized based\n"
                        "on the computed inviscid flow. (You're probably hitting a Goldstein singularity.) Try starting\n"
                        "your XFoil run at a less-aggressive operating point.")
                elif e.returncode == 1:
                    raise RuntimeError(
                        f"Command '{command}' returned non-zero exit status 1.\n"
                        f"This is likely because AeroSandbox does not see XFoil on PATH with the given command.\n"
                        f"Check the logs (`asb.XFoil(..., verbose=True)`) to verify that this is the case, and if so,\n"
                        f"provide the correct path to the XFoil executable in the asb.XFoil constructor via `xfoil_command=`."
                    )
                else:
                    raise e

            ### Parse the polar
            try:
                with open(directory / output_filename) as f:
                    lines = f.readlines()

                title_line = lines[10]
                columns = title_line.split()

                output = {
                    column: []
                    for column in columns
                }

            except FileNotFoundError:
                raise FileNotFoundError(
                    "It appears XFoil didn't produce an output file, probably because it crashed.\n"
                    "Try running with `verbose=True` in the XFoil constructor to see what's going on."
                )

            def str_to_float(s: str) -> float:
                try:
                    return float(s)
                except ValueError:
                    return np.nan

            for line in lines[12:]:
                data = [str_to_float(entry) for entry in line.split()]
                for i in range(len(columns)):
                    output[columns[i]].append(data[i])

            output = {
                k: np.array(v, dtype=float)
                for k, v in output.items()
            }

            return output

    def alpha(self,
              alpha: Union[float, np.ndarray],
              start_at: Union[float, None] = 0,
              ) -> Dict[str, np.ndarray]:
        """
        Execute XFoil at a given angle of attack, or at a sequence of angles of attack.

        Args:

            alpha: The angle of attack [degrees]. Can be either a float or an iterable of floats, such as an array.

            start_at: Chooses whether to split a large sweep into two runs that diverge away from some central value,
            to improve convergence. As an example, if you wanted to sweep from alpha=-20 to alpha=20, you might want
            to instead do two sweeps and stitch them together: 0 to 20, and 0 to -20. `start_at` can be either:

                * None, in which case the alpha inputs are run as a single sequence in the order given.

                * A float that corresponds to an angle of attack (in degrees), in which case the alpha inputs are
                split into two sequences that diverge from the `start_at` value. Successful runs are then sorted by
                `alpha` before returning.

        Returns: A dictionary with the XFoil results. Dictionary values are arrays; they may not be the same shape as
        your input array if some points did not converge.

        """
        alphas = np.array(alpha).reshape(-1)

        if np.length(alphas) > 1:
            if start_at is not None:
                if np.min(alphas) < start_at < np.max(alphas):
                    alphas = np.sort(alphas)
                    alphas_upper = alphas[alphas > start_at]
                    alphas_lower = alphas[alpha <= start_at][::-1]

                    output = self._run_xfoil(
                        "\n".join(
                            [
                                f"a {a}" + ("\nfmom" if self.hinge_point_x is not None else "")
                                for a in alphas_upper
                            ] + [
                                "init"
                            ] + [
                                f"a {a}" + ("\nfmom" if self.hinge_point_x is not None else "")
                                for a in alphas_lower
                            ]
                        )
                    )

                    sort_order = np.argsort(output['alpha'])
                    output = {
                        k: v[sort_order]
                        for k, v in output.items()
                    }
                    return output

        return self._run_xfoil(
            "\n".join([
                f"a {a}" + ("\nfmom" if self.hinge_point_x is not None else "")
                for a in alphas
            ])
        )

    def cl(self,
           cl: Union[float, np.ndarray],
           start_at: Union[float, None] = 0,
           ) -> Dict[str, np.ndarray]:
        """
        Execute XFoil at a given lift coefficient, or at a sequence of lift coefficients.

        Args:
            cl: The lift coefficient [-]. Can be either a float or an iterable of floats, such as an array.

            start_at: Chooses whether to split a large sweep into two runs that diverge away from some central value,
            to improve convergence. As an example, if you wanted to sweep from cl=-1.5 to cl=1.5, you might want to
            instead do two sweeps and stitch them together: 0 to 1.5, and 0 to -1.5. `start_at` can be either:

                * None, in which case the cl inputs are run as a single sequence in the order given.

                * A float that corresponds to an lift coefficient, in which case the cl inputs are
                split into two sequences that diverge from the `start_at` value. Successful runs are then sorted by
                `alpha` before returning.


        Returns: A dictionary with the XFoil results. Dictionary values are arrays; they may not be the same shape as
        your input array if some points did not converge.

        """
        cls = np.array(cl).reshape(-1)

        if np.length(cls) > 1:
            if start_at is not None:
                if np.min(cls) < start_at < np.max(cls):
                    cls = np.sort(cls)
                    cls_upper = cls[cls > start_at]
                    cls_lower = cls[cls <= start_at][::-1]

                    output = self._run_xfoil(
                        "\n".join(
                            [
                                f"cl {c}" + ("\nfmom" if self.hinge_point_x is not None else "")
                                for c in cls_upper
                            ] + [
                                "init"
                            ] + [
                                f"cl {c}" + ("\nfmom" if self.hinge_point_x is not None else "")
                                for c in cls_lower
                            ]
                        )
                    )

                    sort_order = np.argsort(output['alpha'])
                    output = {
                        k: v[sort_order]
                        for k, v in output.items()
                    }
                    return output

        return self._run_xfoil(
            "\n".join([
                f"cl {c}" + ("\nfmom" if self.hinge_point_x is not None else "")
                for c in cls
            ])
        )


if __name__ == '__main__':
    af = Airfoil("naca2412").repanel(n_points_per_side=100)
    # af.coordinates[:, 1] *= 30

    xf = XFoil(
        airfoil=af,
        Re=1e6,
        hinge_point_x=0.75,
        # verbose=True,
    )
    result_at_single_alpha = xf.alpha(5)
    result_at_several_CLs = xf.cl([-0.1, 0.5, 0.7, 0.8, 0.9])
    result_at_multiple_alphas = xf.alpha([3, 5, 60])  # Note: if a result does



from pathlib import Path

_asb_root = Path(__file__).parent

from aerosandbox.common import *
from aerosandbox.optimization import *
from aerosandbox.modeling import *
from aerosandbox.geometry import *
from aerosandbox.atmosphere import *
from aerosandbox.weights import *
from aerosandbox.performance import *
from aerosandbox.dynamics import *
from aerosandbox.aerodynamics import *
from aerosandbox.propulsion import *
from aerosandbox.structures import *

__version__ = "4.0.3"


def docs():
    """
    Opens the AeroSandbox documentation.
    """
    import webbrowser
    webbrowser.open_new(
        "https://github.com/peterdsharpe/AeroSandbox/tree/master/aerosandbox"
    )  # TODO: make this redirect to a hosted ReadTheDocs, or similar.


def run_tests():
    """
    Runs all of the AeroSandbox internal unit tests on this computer.
    """
    try:
        import pytest
    except ModuleNotFoundError:
        raise ModuleNotFoundError("Please install `pytest` (`pip install pytest`) to run AeroSandbox unit tests.")
    import matplotlib.pyplot as plt

    with plt.ion():  # Disable blocking plotting

        pytest.main([str(_asb_root)])



import aerosandbox.numpy as np
from aerosandbox.common import AeroSandboxObject
from abc import ABC, abstractmethod, abstractproperty
from typing import Union, Dict, Tuple, List
from aerosandbox import MassProperties, Opti, OperatingPoint, Atmosphere, Airplane, _asb_root
from aerosandbox.tools.string_formatting import trim_string
import inspect
import copy


class _DynamicsPointMassBaseClass(AeroSandboxObject, ABC):
    @abstractmethod
    def __init__(self,
                 mass_props: MassProperties = None,
                 **state_variables_and_indirect_control_variables,
                 ):
        self.mass_props = MassProperties() if mass_props is None else mass_props
        """
        For each state variable, self.state_var = state_var 
        
        For each indirect control variable, self.indirect_control_var = indirect_control_var
        
        For each control variable, self.control_var = 0
        """

    @abstractproperty
    def state(self) -> Dict[str, Union[float, np.ndarray]]:
        """
        Returns the state variables of this Dynamics instance as a Dict.

        Keys are strings that give the name of the variables.
        Values are the variables themselves.

        This method should look something like:
            >>> {
            >>>     "x_e": self.x_e,
            >>>     "u_e": self.u_e,
            >>>     ...
            >>> }

        """
        pass

    def get_new_instance_with_state(self,
                                    new_state: Union[
                                        Dict[str, Union[float, np.ndarray]],
                                        List, Tuple, np.ndarray
                                    ] = None
                                    ):
        """
        Creates a new instance of this same Dynamics class from the given state.

        Note that any control variables (forces, moments) associated with the previous instance are zeroed.

        Args:
            new_state: The new state to be used for the new instance. Ideally, this is represented as a Dict in identical format to the `state` of a Dynamics instance.

        Returns: A new instance of this same Dynamics class.

        """

        ### Get a list of all the inputs that the class constructor wants to see
        init_signature = inspect.signature(self.__class__.__init__)
        init_args = list(init_signature.parameters.keys())[1:]  # Ignore 'self'

        ### Create a new instance, and give the constructor all the inputs it wants to see (based on values in this instance)
        new_dyn: __class__ = self.__class__(**{
            k: getattr(self, k)
            for k in init_args
        })

        ### Overwrite the state variables in the new instance with those from the input
        new_dyn._set_state(new_state=new_state)

        ### Return the new instance
        return new_dyn

    def _set_state(self,
                   new_state: Union[
                       Dict[str, Union[float, np.ndarray]],
                       List, Tuple, np.ndarray
                   ] = None
                   ):
        """
        Force-overwrites all state variables with a new set (either partial or complete) of state variables.

        Warning: this is *not* the intended public usage of Dynamics instances.
        If you want a new state yourself, you should instantiate a new one either:
            a) manually, or
            b) by using Dynamics.get_new_instance_with_state()

        Hence, this function is meant for PRIVATE use only - be careful how you use this! Especially note that
        control variables (e.g., forces, moments) do not reset to zero.
        """
        ### Set the default parameters
        if new_state is None:
            new_state = {}

        try:  # Assume `value` is a dict-like, with keys
            for key in new_state.keys():  # Overwrite each of the specified state variables
                setattr(self, key, new_state[key])

        except AttributeError:  # Assume it's an iterable that has been sorted.
            self._set_state(
                self.pack_state(new_state))  # Pack the iterable into a dict-like, then do the same thing as above.

    def unpack_state(self,
                     dict_like_state: Dict[str, Union[float, np.ndarray]] = None
                     ) -> Tuple[Union[float, np.ndarray]]:
        """
        'Unpacks' a Dict-like state into an array-like that represents the state of the dynamical system.

        Args:
            dict_like_state: Takes in a dict-like representation of the state.

        Returns: The array representation of the state that you gave.

        """
        if dict_like_state is None:
            dict_like_state = self.state
        return tuple(dict_like_state.values())

    def pack_state(self,
                   array_like_state: Union[List, Tuple, np.ndarray] = None
                   ) -> Dict[str, Union[float, np.ndarray]]:
        """
        'Packs' an array into a Dict that represents the state of the dynamical system.

        Args:
            array_like_state: Takes in an iterable that must have the same number of entries as the state vector of the system.

        Returns: The Dict representation of the state that you gave.

        """
        if array_like_state is None:
            return self.state
        if not len(self.state.keys()) == len(array_like_state):
            raise ValueError(
                "There are a differing number of elements in the `state` variable and the `array_like` you're trying to pack!")
        return {
            k: v
            for k, v in zip(
                self.state.keys(),
                array_like_state
            )
        }

    @abstractproperty
    def control_variables(self) -> Dict[str, Union[float, np.ndarray]]:
        pass

    def __repr__(self) -> str:

        title = f"{self.__class__.__name__} instance:"

        def makeline(k, v):
            name = trim_string(str(k).strip(), length=8).rjust(8)
            item = trim_string(str(v).strip(), length=40).ljust(40)

            line = f"{name}: {item}"

            return line

        state_variables_title = "\tState variables:"

        state_variables = "\n".join([
            "\t\t" + makeline(k, v)
            for k, v in self.state.items()
        ])

        control_variables_title = "\tControl variables:"

        control_variables = "\n".join([
            "\t\t" + makeline(k, v)
            for k, v in self.control_variables.items()
        ])

        return "\n".join([
            title,
            state_variables_title,
            state_variables,
            control_variables_title,
            control_variables
        ])

    def __getitem__(self, index: int):
        """
        Indexes one item from each attribute of a Dynamics instance.
        Returns a new Dynamics instance of the same type.

        Args:
            index: The index that is being called; e.g.,:
                >>> first_dyn = dyn[0]

        Returns: A new Dynamics instance, where each attribute is subscripted at the given value, if possible.

        """

        def get_item_of_attribute(a):
            try:
                return a[index]
            except TypeError as e:  # object is not subscriptable
                return a
            except IndexError as e:  # index out of range
                raise IndexError("A state variable could not be indexed, since the index is out of range!")
            except NotImplementedError as e:
                raise TypeError(f"Indices must be integers or slices, not {index.__class__.__name__}")

        new_instance = self.get_new_instance_with_state()

        for k, v in new_instance.__dict__.items():
            setattr(new_instance, k, get_item_of_attribute(v))

        return new_instance

    def __len__(self):
        """
        Returns the length of the (vectorized) state vector of the system.

        If no elements of the state vector are vectorized, returns 1.
        """
        length = 1
        for v in self.state.values():
            if np.length(v) == 1:
                pass
            elif length == 1:
                length = np.length(v)
            elif length == np.length(v):
                pass
            else:
                raise ValueError("State variables are appear vectorized, but of different lengths!")
        return length

    @abstractmethod
    def state_derivatives(self) -> Dict[str, Union[float, np.ndarray]]:
        """
        A function that returns the derivatives with respect to time of the state specified in the `state` property.

        Should return a Dict with the same keys as the `state` property.
        """
        pass

    def constrain_derivatives(self,
                              opti: Opti,
                              time: np.ndarray,
                              method: str = "midpoint",
                              which: Union[str, List[str], Tuple[str]] = "all",
                              _stacklevel=1,
                              ):
        """
        Applies the relevant state derivative constraints to a given Opti instance.

        Args:

            opti: the AeroSandbox `Opti` instance that constraints should be applied to.

            time: A vector that represents the time at each discrete point. Should be the same length as any
            vectorized state variables in the `state` property of this Dynamics instance.

            method: The discrete integration method to use. See Opti.constrain_derivative() for options.

            which: Which state variables should be we constrain? By default, constrains all of them.

                Options:

                    * "all", which constrains all state variables (default)

                    * A list of strings that are state variable names (i.e., a subset of `dyn.state.keys()`),
                    that gives the names of state variables to be constrained.

            _stacklevel: Optional and advanced, purely used for debugging. Allows users to correctly track where
            constraints are declared in the event that they are subclassing `aerosandbox.Opti`. Modifies the
            stacklevel of the declaration tracked, which is then presented using
            `aerosandbox.Opti.variable_declaration()` and `aerosandbox.Opti.constraint_declaration()`.

        Returns:

        """
        if which == "all":
            which = self.state.keys()

        state_derivatives = self.state_derivatives()

        for state_var_name in which:

            # If a state derivative has a None value, skip it.
            if state_derivatives[state_var_name] is None:
                continue

            # Try to constrain the derivative
            try:
                opti.constrain_derivative(
                    derivative=state_derivatives[state_var_name],
                    variable=self.state[state_var_name],
                    with_respect_to=time,
                    method=method,
                    _stacklevel=_stacklevel + 1
                )
            except KeyError:
                raise ValueError(f"This dynamics instance does not have a state named '{state_var_name}'!")
            except Exception as e:
                raise ValueError(f"Error while constraining state variable '{state_var_name}': \n{e}")

    @abstractmethod
    def convert_axes(self,
                     x_from: float,
                     y_from: float,
                     z_from: float,
                     from_axes: str,
                     to_axes: str,
                     ) -> Tuple[float, float, float]:
        """
        Converts a vector [x_from, y_from, z_from], as given in the `from_axes` frame, to an equivalent vector [x_to,
        y_to, z_to], as given in the `to_axes` frame.

        Identical to OperatingPoint.convert_axes(), but adds in "earth" as a valid axis frame. For more documentation,
        see the docstring of OperatingPoint.convert_axes().

        Both `from_axes` and `to_axes` should be a string, one of:
                * "geometry"
                * "body"
                * "wind"
                * "stability"
                * "earth"

        Args:
                x_from: x-component of the vector, in `from_axes` frame.

                y_from: y-component of the vector, in `from_axes` frame.

                z_from: z-component of the vector, in `from_axes` frame.

                from_axes: The axes to convert from. See above for options.

                to_axes: The axes to convert to. See above for options.


        Returns: The x-, y-, and z-components of the vector, in `to_axes` frame. Given as a tuple.

        """
        pass

    @abstractmethod
    def add_force(self,
                  Fx: Union[float, np.ndarray] = 0,
                  Fy: Union[float, np.ndarray] = 0,
                  Fz: Union[float, np.ndarray] = 0,
                  axes: str = "wind",
                  ) -> None:
        """
        Adds a force (in whichever axis system you choose) to this Dynamics instance.

        Args:
            Fx: Force in the x-direction in the axis system chosen. [N]

            Fy: Force in the y-direction in the axis system chosen. [N]

            Fz: Force in the z-direction in the axis system chosen. [N]

            axes: The axis system that the specified force is in. One of:
                * "geometry"
                * "body"
                * "wind"
                * "stability"
                * "earth"

        Returns: None (in-place)

        """
        pass

    def add_gravity_force(self,
                          g=9.81
                          ) -> None:
        """
        In-place modifies the forces associated with this Dynamics instance: adds a force in the -z direction,
        equal to the weight of the aircraft.

        Args:
            g: The gravitational acceleration. [m/s^2]

        Returns: None (in-place)
        """
        self.add_force(
            Fz=self.mass_props.mass * g,
            axes="earth",
        )

    @property
    def op_point(self):
        """
        Returns an OperatingPoint object that represents the current state of the dynamics instance.

        This OperatingPoint object is effectively a subset of the state variables, and is used to compute aerodynamic
        forces and moments.
        """
        return OperatingPoint(
            atmosphere=Atmosphere(altitude=self.altitude),
            velocity=self.speed,
            alpha=self.alpha,
            beta=self.beta,
            p=0,
            q=0,
            r=0,
        )

    def draw(self,
             vehicle_model: Airplane = None,
             backend: str = "pyvista",
             draw_axes: bool = True,
             scale_vehicle_model: Union[float, None] = None,
             n_vehicles_to_draw: int = 10,
             cg_axes: str = "geometry",
             show: bool = True,
             ):
        if backend == "pyvista":
            import pyvista as pv
            import aerosandbox.tools.pretty_plots as p

            if vehicle_model is None:
                default_vehicle_stl = _asb_root / "dynamics/visualization/default_assets/yf23.stl"
                vehicle_model = pv.read(str(default_vehicle_stl))
            elif isinstance(vehicle_model, pv.PolyData):
                pass
            elif isinstance(vehicle_model, Airplane):
                vehicle_model = vehicle_model.draw(
                    backend="pyvista",
                    show=False
                )
                vehicle_model.rotate_y(180, inplace=True)  # Rotate from geometry axes to body axes.
            elif isinstance(vehicle_model, str):  # Interpret the string as a filepath to a .stl or similar
                try:
                    pv.read(filename=vehicle_model)
                except:
                    raise ValueError("Could not parse `vehicle_model`!")
            else:
                raise TypeError("`vehicle_model` should be an Airplane or PolyData object.")

            x_e = np.array(self.x_e)
            y_e = np.array(self.y_e)
            z_e = np.array(self.z_e)
            if np.length(x_e) == 1:
                x_e = x_e * np.ones(len(self))
            if np.length(y_e) == 1:
                y_e = y_e * np.ones(len(self))
            if np.length(z_e) == 1:
                z_e = z_e * np.ones(len(self))

            if scale_vehicle_model is None:
                trajectory_bounds = np.array([
                    [x_e.min(), x_e.max()],
                    [y_e.min(), y_e.max()],
                    [z_e.min(), z_e.max()],
                ])
                trajectory_size = np.max(np.diff(trajectory_bounds, axis=1))

                vehicle_bounds = np.array(vehicle_model.bounds).reshape((3, 2))
                vehicle_size = np.max(np.diff(vehicle_bounds, axis=1))

                scale_vehicle_model = 0.1 * trajectory_size / vehicle_size

            ### Initialize the plotter
            plotter = pv.Plotter()

            # Set the window title
            title = "ASB Dynamics"
            addenda = []
            if scale_vehicle_model != 1:
                addenda.append(f"Vehicle drawn at {scale_vehicle_model:.2g}x scale")
            addenda.append(f"{self.__class__.__name__} Engine")
            if len(addenda) != 0:
                title = title + f" ({'; '.join(addenda)})"
            plotter.title = title

            # Draw axes and grid
            plotter.add_axes()
            plotter.show_grid(color='gray')

            ### Draw the vehicle
            for i in np.unique(
                    np.round(
                        np.linspace(0, len(self) - 1, n_vehicles_to_draw)
                    )
            ).astype(np.int64):
                dyn = self[i]
                try:
                    phi = dyn.phi
                except AttributeError:
                    phi = dyn.bank
                try:
                    theta = dyn.theta
                except AttributeError:
                    theta = dyn.gamma
                try:
                    psi = dyn.psi
                except AttributeError:
                    psi = dyn.track

                x_cg_b, y_cg_b, z_cg_b = dyn.convert_axes(
                    dyn.mass_props.x_cg,
                    dyn.mass_props.y_cg,
                    dyn.mass_props.z_cg,
                    from_axes=cg_axes,
                    to_axes="body"
                )

                this_vehicle = copy.deepcopy(vehicle_model)
                this_vehicle.translate([
                    -x_cg_b,
                    -y_cg_b,
                    -z_cg_b,
                ], inplace=True)
                this_vehicle.points *= scale_vehicle_model
                this_vehicle.rotate_x(np.degrees(phi), inplace=True)
                this_vehicle.rotate_y(np.degrees(theta), inplace=True)
                this_vehicle.rotate_z(np.degrees(psi), inplace=True)
                this_vehicle.translate([
                    dyn.x_e,
                    dyn.y_e,
                    dyn.z_e,
                ], inplace=True)
                plotter.add_mesh(
                    this_vehicle,
                )
                if draw_axes:
                    rot = np.rotation_matrix_from_euler_angles(phi, theta, psi)
                    axes_scale = 0.5 * np.max(
                        np.diff(
                            np.array(this_vehicle.bounds).reshape((3, -1)),
                            axis=1
                        )
                    )
                    origin = np.array([
                        dyn.x_e,
                        dyn.y_e,
                        dyn.z_e,
                    ])
                    for i, c in enumerate(["r", "g", "b"]):
                        plotter.add_mesh(
                            pv.Spline(np.array([
                                origin,
                                origin + rot[:, i] * axes_scale
                            ])),
                            color=c,
                            line_width=2.5,
                        )

            for i in range(len(self)):
                ### Draw the trajectory line

                polyline = pv.Spline(np.array([x_e, y_e, z_e]).T)
                plotter.add_mesh(
                    polyline,
                    color=p.adjust_lightness(p.palettes["categorical"][0], 1.2),
                    line_width=3,
                )

            ### Finalize the plotter
            plotter.camera.up = (0, 0, -1)
            plotter.camera.Azimuth(90)
            plotter.camera.Elevation(60)
            if show:
                plotter.show()
            return plotter

    @property
    def altitude(self):
        return -self.z_e

    @property
    def translational_kinetic_energy(self) -> float:
        """
        Computes the kinetic energy [J] from translational motion.

        KE = 0.5 * m * v^2

        Returns:
            Kinetic energy [J]
        """
        return 0.5 * self.mass_props.mass * self.speed ** 2

    @property
    def rotational_kinetic_energy(self) -> float:
        """
        Computes the kinetic energy [J] from rotational motion.

        KE = 0.5 * I * w^2

        Returns:
            Kinetic energy [J]
        """
        return 0.5 * (
                self.mass_props.Ixx * self.p ** 2 +
                self.mass_props.Iyy * self.q ** 2 +
                self.mass_props.Izz * self.r ** 2
        )

    @property
    def kinetic_energy(self):
        """
        Computes the kinetic energy [J] from translational and rotational motion.

        KE = 0.5 * m * v^2 + 0.5 * I * w^2

        Returns:
            Kinetic energy [J]
        """
        return self.translational_kinetic_energy + self.rotational_kinetic_energy

    @property
    def potential_energy(self,
                         g: float = 9.81
                         ):
        """
        Computes the potential energy [J] from gravity.

        PE = mgh

        Args:
            g: Acceleration due to gravity [m/s^2]

        Returns:
            Potential energy [J]
        """
        return self.mass_props.mass * g * self.altitude



from pathlib import Path
import json


def convert_ipynb_to_py(
        input_file: Path,
        output_file: Path,
) -> None:
    """
    Reads an input Jupyter notebook (.ipynb) and converts it to a Python file (.py)

    Tried using `jupyter nbconvert`, but that is SO SLOW, like 3 seconds per notebook! It's just json parsing,
    this should *not* take more than a few milliseconds - come on, Jupyter!

    Args:
        input_file: File path
        output_file: File path

    Returns: None

    """
    with open(input_file, "r") as f:
        ipynb_contents = json.load(f)
    with open(output_file, "w+") as f:
        for cell in ipynb_contents['cells']:
            if cell['cell_type'] == "code":
                f.writelines(cell['source'])
                f.write("\n")



import aerosandbox as asb
import aerosandbox.numpy as np
import aerosandbox.tools.units as u
from .raymer_fudge_factors import advanced_composites


# From Raymer: "Aircraft Design: A Conceptual Approach", 5th Ed.
# Section 15.3.3: General Aviation Weights

def mass_wing(
        wing: asb.Wing,
        design_mass_TOGW: float,
        ultimate_load_factor: float,
        mass_fuel_in_wing: float,
        cruise_op_point: asb.OperatingPoint,
        use_advanced_composites: bool = False,
) -> float:
    """
    Computes the mass of a wing of a general aviation aircraft, according to Raymer's Aircraft Design: A Conceptual
    Approach.

    Note: Torenbeek's wing mass model is likely more accurate; see `mass_wing()` in `torenbeek_weights.py` (same
    directory).

    Args:
        wing: The wing object.

        design_mass_TOGW: The design takeoff gross weight of the entire aircraft [kg].

        ultimate_load_factor: The ultimate load factor of the aircraft.

        mass_fuel_in_wing: The mass of fuel in the wing [kg]. If there is no fuel in the wing, set this to 0.

            Note: Model extrapolates strangely for infinitesimally-small-but-nonzero fuel masses; don't let an
            optimizer land here.

        cruise_op_point: The cruise operating point of the aircraft.

        use_advanced_composites: Whether to use advanced composites for the wing. If True, the wing mass is modified
        accordingly.

    Returns: The mass of the wing [kg].

    """
    try:
        fuel_is_in_wing = bool(mass_fuel_in_wing > 0)
    except RuntimeError:
        fuel_is_in_wing = True

    if fuel_is_in_wing:
        fuel_weight_factor = np.softmax(
            (mass_fuel_in_wing / u.lbm) ** 0.0035,
            1,
            hardness=1000
        )
    else:
        fuel_weight_factor = 1

    airfoil_thicknesses = [
        xsec.airfoil.max_thickness()
        for xsec in wing.xsecs
    ]

    airfoil_t_over_c = np.min(airfoil_thicknesses)

    cos_sweep = np.cosd(wing.mean_sweep_angle())

    return (
            0.036 *
            (wing.area('planform') / u.foot ** 2) ** 0.758 *
            fuel_weight_factor *
            (wing.aspect_ratio() / cos_sweep ** 2) ** 0.6 *
            (cruise_op_point.dynamic_pressure() / u.psf) ** 0.006 *
            wing.taper_ratio() ** 0.04 *
            (100 * airfoil_t_over_c / cos_sweep) ** -0.3 *
            (design_mass_TOGW / u.lbm * ultimate_load_factor) ** 0.49 *
            (advanced_composites["wing"] if use_advanced_composites else 1)
    ) * u.lbm


def mass_hstab(
        hstab: asb.Wing,
        design_mass_TOGW: float,
        ultimate_load_factor: float,
        cruise_op_point: asb.OperatingPoint,
        use_advanced_composites: bool = False,
) -> float:
    """
    Computes the mass of a horizontal stabilizer of a general aviation aircraft, according to Raymer's Aircraft Design:
    A Conceptual Approach.

    Args:
        hstab: The horizontal stabilizer object.

        design_mass_TOGW: The design takeoff gross weight of the entire aircraft [kg].

        ultimate_load_factor: The ultimate load factor of the aircraft.

        cruise_op_point: The cruise operating point of the aircraft.

        use_advanced_composites: Whether to use advanced composites for the horizontal stabilizer. If True, the
        hstab mass is modified accordingly.

    Returns: The mass of the horizontal stabilizer [kg].
    """
    airfoil_thicknesses = [
        xsec.airfoil.max_thickness()
        for xsec in hstab.xsecs
    ]

    airfoil_t_over_c = np.min(airfoil_thicknesses)

    cos_sweep = np.cosd(hstab.mean_sweep_angle())

    return (
            0.016 *
            (design_mass_TOGW / u.lbm * ultimate_load_factor) ** 0.414 *
            (cruise_op_point.dynamic_pressure() / u.psf) ** 0.168 *
            (hstab.area('planform') / u.foot ** 2) ** 0.896 *
            (100 * airfoil_t_over_c / cos_sweep) ** -0.12 *
            (hstab.aspect_ratio() / cos_sweep ** 2) ** 0.043 *
            hstab.taper_ratio() ** -0.02 *
            (advanced_composites["tails"] if use_advanced_composites else 1)
    ) * u.lbm


def mass_vstab(
        vstab: asb.Wing,
        design_mass_TOGW: float,
        ultimate_load_factor: float,
        cruise_op_point: asb.OperatingPoint,
        is_t_tail: bool = False,
        use_advanced_composites: bool = False,
) -> float:
    """
    Computes the mass of a vertical stabilizer of a general aviation aircraft, according to Raymer's Aircraft Design:
    A Conceptual Approach.

    Args:
        vstab: The vertical stabilizer object.

        design_mass_TOGW: The design takeoff gross weight of the entire aircraft [kg].

        ultimate_load_factor: The ultimate load factor of the aircraft.

        cruise_op_point: The cruise operating point of the aircraft.

        is_t_tail: Whether the aircraft is a T-tail or not.

        use_advanced_composites: Whether to use advanced composites for the vertical stabilizer. If True, the vstab
        mass is modified accordingly.

    Returns: The mass of the vertical stabilizer [kg].
    """
    airfoil_thicknesses = [
        xsec.airfoil.max_thickness()
        for xsec in vstab.xsecs
    ]

    airfoil_t_over_c = np.min(airfoil_thicknesses)

    cos_sweep = np.cosd(vstab.mean_sweep_angle())

    return (
            0.073 *
            (1 + (0.2 if is_t_tail else 0)) *
            (design_mass_TOGW / u.lbm * ultimate_load_factor) ** 0.376 *
            (cruise_op_point.dynamic_pressure() / u.psf) ** 0.122 *
            (vstab.area('planform') / u.foot ** 2) ** 0.876 *
            (100 * airfoil_t_over_c / cos_sweep) ** -0.49 *
            (vstab.aspect_ratio() / cos_sweep ** 2) ** 0.357 *
            vstab.taper_ratio() ** 0.039 *
            (advanced_composites["tails"] if use_advanced_composites else 1)
    ) * u.lbm


def mass_fuselage(
        fuselage: asb.Fuselage,
        design_mass_TOGW: float,
        ultimate_load_factor: float,
        L_over_D: float,
        cruise_op_point: asb.OperatingPoint,
        wing_to_tail_distance: float,
        pressure_differential: float = 0.0,
        use_advanced_composites: bool = False,
) -> float:
    """
    Computes the mass of a fuselage of a general aviation aircraft, according to Raymer's Aircraft Design: A Conceptual
    Approach.

    Args:
        fuselage: The fuselage object.

        design_mass_TOGW: The design takeoff gross weight of the entire aircraft [kg].

        ultimate_load_factor: The ultimate load factor of the aircraft.

        L_over_D: The lift-to-drag ratio of the aircraft in cruise.

        cruise_op_point: The cruise operating point of the aircraft.

        wing_to_tail_distance: The distance between the wing root-quarter-chord-point and the tail
        root-quarter-chord-point of the aircraft [m].

        pressure_differential: The absolute value of the pressure differential across the fuselage [Pa].

        use_advanced_composites: Whether to use advanced composites for the fuselage. If True, the fuselage mass is
        modified accordingly.

    Returns: The mass of the fuselage [kg].
    """

    mass_fuselage_without_pressurization = (
                                                   0.052 *
                                                   (fuselage.area_wetted() / u.foot ** 2) ** 1.086 *
                                                   (design_mass_TOGW / u.lbm * ultimate_load_factor) ** 0.177 *
                                                   (wing_to_tail_distance / u.foot) ** -0.051 *
                                                   (L_over_D) ** -0.072 *
                                                   (cruise_op_point.dynamic_pressure() / u.psf) ** 0.241 *
                                                   (advanced_composites["fuselage/nacelle"]
                                                    if use_advanced_composites else 1)
                                           ) * u.lbm

    mass_pressurization_components = (
                                             11.9 *
                                             (
                                                     fuselage.volume() / u.foot ** 3 *
                                                     pressure_differential / u.psi
                                             ) ** 0.271
                                     ) * u.lbm

    return (
            mass_fuselage_without_pressurization +
            mass_pressurization_components
    )


def mass_main_landing_gear(
        main_gear_length: float,
        design_mass_TOGW: float,
        n_gear: int = 2,
        is_retractable: bool = True,
        use_advanced_composites: bool = False,
) -> float:
    """
    Computes the mass of the main landing gear of a general aviation aircraft, according to Raymer's Aircraft Design:
    A Conceptual Approach.

    Args:
        main_gear_length: The length of the main landing gear [m].

        design_mass_TOGW: The design takeoff gross weight of the entire aircraft [kg].

        n_gear: The number of main landing gear.

        is_retractable: Whether the main landing gear is retractable or not.

        use_advanced_composites: Whether to use advanced composites for the main landing gear. If True, the main
        landing gear mass is modified accordingly.

    Returns: The mass of the main landing gear [kg].
    """

    ultimate_landing_load_factor = n_gear * 1.5

    return (
            0.095 *
            (ultimate_landing_load_factor * design_mass_TOGW / u.lbm) ** 0.768 *
            (main_gear_length / u.foot / 12) ** 0.409 *
            (advanced_composites["landing_gear"] if use_advanced_composites else 1) *
            (((5.7 - 1.4 / 2) / 5.7) if not is_retractable else 1)  # derived from Raymer Section 15.2 and 15.3.3 together.
    ) * u.lbm


def mass_nose_landing_gear(
        nose_gear_length: float,
        design_mass_TOGW: float,
        n_gear: int = 1,
        is_retractable: bool = True,
        use_advanced_composites: bool = False,
) -> float:
    """
    Computes the mass of the nose landing gear of a general aviation aircraft, according to Raymer's Aircraft Design:
    A Conceptual Approach.

    Args:
        nose_gear_length: The length of the nose landing gear [m].

        design_mass_TOGW: The design takeoff gross weight of the entire aircraft [kg].

        n_gear: The number of nose landing gear.

        is_retractable: Whether the nose landing gear is retractable or not.

        use_advanced_composites: Whether to use advanced composites for the nose landing gear. If True, the nose
        landing gear mass is modified accordingly.

    Returns: The mass of the nose landing gear [kg].
    """

    ultimate_landing_load_factor = n_gear * 1.5

    return (
            0.125 *
            (ultimate_landing_load_factor * design_mass_TOGW / u.lbm) ** 0.566 *
            (nose_gear_length / u.foot / 12) ** 0.845 *
            (advanced_composites["landing_gear"] if use_advanced_composites else 1) *
            (((5.7 - 1.4 / 2) / 5.7) if not is_retractable else 1)  # derived from Raymer Section 15.2 and 15.3.3 together.
    ) * u.lbm


def mass_engines_installed(
        n_engines: int,
        mass_per_engine: float,
) -> float:
    """
    Computes the mass of the engines installed on a general aviation aircraft, according to Raymer's Aircraft Design:
    A Conceptual Approach. Includes propellers and engine mounts.

    Args:
        n_engines: The number of engines installed on the aircraft.

        mass_per_engine: The mass of a single engine [kg].

    Returns: The mass of the engines installed on the aircraft [kg].
    """
    return (
            2.575 *
            (mass_per_engine / u.lbm) ** 0.922 *
            n_engines
    ) * u.lbm


def mass_fuel_system(
        fuel_volume: float,
        n_tanks: int,
        n_engines: int,
        fraction_in_integral_tanks: float = 0.5,
) -> float:
    """
    Computes the mass of the fuel system (e.g., tanks, pumps, but not the fuel itself) for a general aviation
    aircraft, according to Raymer's Aircraft Design: A Conceptual Approach.

    Args:
        fuel_volume: The volume of fuel in the aircraft [m^3].

        n_tanks: The number of fuel tanks in the aircraft.

        n_engines: The number of engines in the aircraft.

        fraction_in_integral_tanks: The fraction of the fuel volume that is in integral tanks, as opposed to
        protected tanks.

    Returns: The mass of the fuel system [kg].
    """
    return (
            2.49 *
            (fuel_volume / u.gallon) ** 0.726 *
            (1 + fraction_in_integral_tanks) ** -0.363 *
            n_tanks ** 0.242 *
            n_engines ** 0.157
    ) * u.lbm


def mass_flight_controls(
        airplane: asb.Airplane,
        design_mass_TOGW: float,
        ultimate_load_factor: float,
        fuselage: asb.Fuselage = None,
        main_wing: asb.Wing = None,
) -> float:
    """
    Computes the mass of the flight controls for a general aviation aircraft, according to Raymer's Aircraft Design:
    A Conceptual Approach.

    Args:
        airplane: The airplane for which to compute the flight controls mass.

        design_mass_TOGW: The design takeoff gross weight of the entire aircraft [kg].

        ultimate_load_factor: The ultimate load factor of the aircraft.

        fuselage: The fuselage to use for computing the flight controls mass. If fuselage is None, or if there are no
        fuselages in the airplane object, the flight controls mass will be computed without a fuselage.

        main_wing: The main wing to use for computing the flight controls mass. If main_wing is None, or if there are
        no wings in the airplane object, the flight controls mass will be computed without a main wing.

    Returns: The mass of the flight controls [kg].
    """

    ### Handle the fuselage argument and get the fuselage length factor
    if fuselage is None:
        if len(airplane.fuselages) == 0:
            pass
        elif len(airplane.fuselages) == 1:
            fuselage = airplane.fuselages[0]
        else:
            raise ValueError('More than one fuselage is present in the airplane. Please specify which fuselage to use '
                             'for computing flight control system mass.')

    if fuselage is not None:
        fuselage_length_factor = (fuselage.length() / u.foot) ** 1.536
    else:
        fuselage_length_factor = 1

    ### Handle the main wing argument and get the wing span factor
    if main_wing is None:
        if len(airplane.wings) == 0:
            pass
        elif len(airplane.wings) == 1:
            main_wing = airplane.wings[0]
        else:
            raise ValueError('More than one wing is present in the airplane. Please specify which wing is the main'
                             'wing using the `main_wing` argument.')

    if main_wing is not None:
        wing_span_factor = (main_wing.span() / u.foot) ** 0.371
    else:
        wing_span_factor = 1

    # ### Compute how many functions the control surfaces are performing (e.g., aileron, elevator, flap, rudder, etc.)
    # N_functions_performed_by_controls = 0
    # for wing in airplane.wings:
    #     N_functions_performed_by_controls += len(wing.get_control_surface_names())
    #
    # ### Compute the control surface area
    # control_surface_area = 0
    # for wing in airplane.wings:
    #     control_surface_area += wing.control_surface_area()

    return (
            0.053 *
            fuselage_length_factor *
            wing_span_factor *
            (design_mass_TOGW / u.lbm * ultimate_load_factor * 1e-4) ** 0.80
    ) * u.lbm


def mass_hydraulics(
        fuselage_width: float,
        cruise_op_point: asb.OperatingPoint,
) -> float:
    """
    Computes the mass of the hydraulics for a general aviation aircraft, according to Raymer's Aircraft Design:
    A Conceptual Approach.

    Args:
        fuselage_width: The width of the fuselage [m].

        cruise_op_point: The cruise operating point of the aircraft.

    Returns: The mass of the hydraulics [kg].
    """
    mach = cruise_op_point.mach()

    K_h = 0.16472092991402892 * mach ** 0.8327375101470056
    # This is a curve fit to a few points that Raymer gives in his book. The points are:
    # {
    #     0.1 : 0.013,
    #     0.25: 0.05,
    #     0.5 : 0.11,
    #     0.75: 0.12
    # }
    # where the first column is the Mach number and the second column is the K_h value.
    # These are described as:
    #
    # "0.05 for low subsonic with hydraulics for brakes and retracts only; 0.11 for medium subsonic with hydraulics
    # for flaps; 0.12 for high subsonic with hydraulic flight controls; 0.013 for light plane with hydraulic brakes
    # only (and use M=0.1)"

    return (
            K_h *
            (fuselage_width / u.foot) ** 0.8 *
            mach ** 0.5
    ) * u.lbm


def mass_avionics(
        mass_uninstalled_avionics: float,
) -> float:
    """
    Computes the mass of the avionics for a general aviation aircraft, according to Raymer's Aircraft Design: A
    Conceptual Approach.

    Args:
        mass_uninstalled_avionics: The mass of the avionics, before installation [kg].

    Returns: The mass of the avionics, as installed [kg].
    """
    return (
            2.117 *
            (mass_uninstalled_avionics / u.lbm) ** 0.933
    ) * u.lbm


def mass_electrical(
        fuel_system_mass: float,
        avionics_mass: float,
) -> float:
    """
    Computes the mass of the electrical system for a general aviation aircraft, according to Raymer's Aircraft Design:
    A Conceptual Approach.

    Args:
        fuel_system_mass: The mass of the fuel system [kg].

        avionics_mass: The mass of the avionics [kg].

    Returns: The mass of the electrical system [kg].
    """

    fuel_and_avionics_masses = fuel_system_mass + avionics_mass

    return (
        12.57 *
        (fuel_and_avionics_masses / u.lbm) ** 0.51
    ) * u.lbm


def mass_air_conditioning_and_anti_ice(
        design_mass_TOGW: float,
        n_crew: int,
        n_pax: int,
        mass_avionics: float,
        cruise_op_point: asb.OperatingPoint,
):
    """
    Computes the mass of the air conditioning and anti-ice system for a general aviation aircraft, according to
    Raymer's Aircraft Design: A Conceptual Approach.

    Args:
        design_mass_TOGW: The design takeoff gross weight of the entire airplane [kg].

        n_crew: The number of crew members.

        n_pax: The number of passengers.

        mass_avionics: The mass of the avionics [kg].

        cruise_op_point: The cruise operating point of the aircraft.

    Returns: The mass of the air conditioning and anti-ice system [kg].
    """
    mach = cruise_op_point.mach()

    return (
            0.265 *
            (design_mass_TOGW / u.lbm) ** 0.52 *
            (n_crew + n_pax) ** 0.68 *
            (mass_avionics / u.lbm) ** 0.17 *
            mach ** 0.08
    ) * u.lbm


def mass_furnishings(
        design_mass_TOGW: float,
):
    """
    Computes the mass of the furnishings for a general aviation aircraft, according to Raymer's Aircraft Design: A
    Conceptual Approach.

    Args:
        design_mass_TOGW: The design takeoff gross weight of the entire airplane [kg].

    Returns: The mass of the furnishings [kg].
    """
    return np.softmax(
        0.0582 * design_mass_TOGW - 65 * u.lbm,
        0,
        softness=10 * u.lbm,
    )



from aerosandbox.tools import inspect_tools
from typing import Union, Tuple, List
import aerosandbox.numpy as np
import warnings


def qp(
        *args: Tuple[Union[np.ndarray, List]],
        backend="plotly",
        show=True,
        plotly_renderer: Union[str, None] = "browser",
        orthographic=True,
        stacklevel=1
) -> None:
    """
    Quickly plots ("QP") a 1D, 2D, or 3D dataset as a line plot with markers. Useful for exploratory data analysis.

    Example:

        >>> import aerosandbox.numpy as np
        >>>
        >>> x = np.linspace(0, 10)
        >>> y = x ** 2
        >>> z = np.sin(x)
        >>> qp(x, y, z)

    Args:
        *args: The arguments that you want to plot. You can provide 1, 2, or 3 arrays, all of which should be 1D and of the same length.

        backend: The backend to use. Current options:
            * "plotly"

        show: A boolean of whether or not to show the plot.

        plotly_renderer: A string of what to use as the Plotly renderer. If you don't want to overwrite a default that you've already set, set this variable to None.

        orthographic: A boolean of whether or not to use an orthographic (rather than persepctive) projection when viewing 3D plots.

        stacklevel: (Advanced) Choose the level of the stack that you want to retrieve plot labels at. Higher
        integers will get you higher (i.e., more end-user-facing) in the stack. Same behaviour as the `stacklevel`
        argument in warnings.warn().

    Returns: None (in-place)

    """
    arg_values = args
    n_dimensions = len(arg_values)  # dimensionality

    ##### This section serves to try to retrieve appropriate axis labels for the plot.
    ### First, set some defaults.
    arg_names = []
    if n_dimensions >= 1:
        arg_names += ["x"]
        if n_dimensions >= 2:
            arg_names += ["y"]
            if n_dimensions >= 3:
                arg_names += ["z"]
                if n_dimensions >= 4:
                    arg_names += [
                        f"Dim. {i}"
                        for i in range(4, n_dimensions + 1)
                    ]
    title = "QuickPlot"
    try:
        ### This is some interesting and tricky code here: retrieves the source code of where qp() was called, as a string.
        caller_source_code = inspect_tools.get_caller_source_code(stacklevel=stacklevel + 1)
        try:
            parsed_arg_names = inspect_tools.get_function_argument_names_from_source_code(caller_source_code)
            title = "QuickPlot: " + " vs. ".join(parsed_arg_names)
            if len(parsed_arg_names) == n_dimensions:
                arg_names = parsed_arg_names
            else:
                warnings.warn(
                    f"Couldn't parse QuickPlot call signature (dimension mismatch):\n\n{caller_source_code}",
                    stacklevel=2,
                )
        except ValueError as e:
            warnings.warn(
                f"Couldn't parse QuickPlot call signature (invalid source code):\n\n{caller_source_code}",
                stacklevel=2,
            )
    except FileNotFoundError:
        warnings.warn(
            f"Couldn't parse QuickPlot call signature (missing filepath).",
            stacklevel=2,
        )

    ##### Do the plotting:

    if backend == "plotly":
        import plotly.express as px
        import plotly.graph_objects as go

        mode = "markers+lines"
        marker_dict = dict(
            size=5 if n_dimensions != 3 else 2,
            line=dict(
                width=0
            )
        )

        if n_dimensions == 1:
            fig = go.Figure(
                data=go.Scatter(
                    y=arg_values[0],
                    mode=mode,
                    marker=marker_dict
                )
            )
            fig.update_layout(
                title=title,
                xaxis_title="Array index #",
                yaxis_title=arg_names[0]
            )
        elif n_dimensions == 2:
            fig = go.Figure(
                data=go.Scatter(
                    x=arg_values[0],
                    y=arg_values[1],
                    mode=mode,
                    marker=marker_dict
                )
            )
            fig.update_layout(
                title=title,
                xaxis_title=arg_names[0],
                yaxis_title=arg_names[1]
            )
        elif n_dimensions == 3:
            fig = go.Figure(
                data=go.Scatter3d(
                    x=arg_values[0],
                    y=arg_values[1],
                    z=arg_values[2],
                    mode=mode,
                    marker=marker_dict
                ),
            )
            fig.update_layout(
                title=title,
                scene=dict(
                    xaxis_title=arg_names[0],
                    yaxis_title=arg_names[1],
                    zaxis_title=arg_names[2],
                )
            )
        else:
            raise ValueError("Too many inputs to plot!")
        # fig.data[0].update(mode='markers+lines')
        if orthographic:
            fig.layout.scene.camera.projection.type = "orthographic"
        if show:
            fig.show(
                renderer=plotly_renderer
            )
    else:
        raise ValueError("Bad value of `backend`!")


if __name__ == '__main__':
    import aerosandbox.numpy as np

    x = np.linspace(0, 10, 100)
    y = x ** 2
    z = np.sin(y)
    qp(x)
    qp(x, y)
    qp(x, y, z)



import aerosandbox.numpy as np
from typing import Tuple

"""
Documentation of (points, faces) standard format, which is an unstructured mesh format:

Meshes are given here in the common (points, faces) format. In this format, `points` is a Nx3 array, where each row 
gives the 3D coordinates of a vertex in the mesh. Entries into this array are floating-point, generally speaking.

`faces` is a Mx3 array in the case of a triangular mesh, or a Mx4 array in the case of a quadrilateral mesh. Each row 
in this array represents a face. The entries in each row are integers that correspond to the index of `points` where 
the vertex locations of that face are found. 

"""


def stack_meshes(
        *meshes: Tuple[Tuple[np.ndarray, np.ndarray]]
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Takes in a series of tuples (points, faces) and merges them into a single tuple (points, faces). All (points,
    faces) tuples are meshes given in standard format.

    Args:
        *meshes: Any number of mesh tuples in standard (points, faces) format.

    Returns: Points and faces of the combined mesh. Standard unstructured mesh format: A tuple of `points` and
    `faces`, where:

        * `points` is a `n x 3` array of points, where `n` is the number of points in the mesh.

        * `faces` is a `m x 3` array of faces if `method` is "tri", or a `m x 4` array of faces if `method` is "quad".

            * Each row of `faces` is a list of indices into `points`, which specifies a face.

    """
    if len(meshes) == 1:
        return meshes[0]
    elif len(meshes) == 2:
        points1, faces1 = meshes[0]
        points2, faces2 = meshes[1]

        faces2 = faces2 + len(points1)

        points = np.concatenate((points1, points2))
        faces = np.concatenate((faces1, faces2))

        return points, faces
    else:
        points, faces = stack_meshes(
            meshes[0],
            meshes[1]
        )
        return stack_meshes(
            (points, faces),
            *meshes[2:]
        )


def convert_mesh_to_polydata_format(
        points: np.ndarray,
        faces: np.ndarray
) -> Tuple[np.ndarray, np.ndarray]:
    """
    PyVista uses a slightly different convention for the standard (points, faces) format as described above. They
    give `faces` as a single 1D vector of roughly length (M*3), or (M*4) in the case of quadrilateral meshing.
    Basically, the mesh displayer goes down the `faces` array, and when it sees a number N, it interprets that as the
    number of vertices in the following face. Then, the next N entries are interpreted as integer references to the
    vertices of the face.

    This has the benefit of allowing for mixed tri/quad meshes.

    Args:
        points: `points` array of the original standard-format mesh
        faces: `faces` array of the original standard-format mesh

    Returns:

        (points, faces), except that `faces` is now in a pyvista.PolyData compatible format.

    """
    faces = [
        [len(face), *face]
        for face in faces
    ]
    faces = np.array(faces)
    faces = np.reshape(faces, -1)
    return points, faces



import aerosandbox as asb
import aerosandbox.numpy as np

import matplotlib.pyplot as plt
import aerosandbox.tools.pretty_plots as p

fuselage = asb.Fuselage(
    xsecs=[
        asb.FuselageXSec(
            xyz_c=[xi, 0, 0],
            radius=asb.Airfoil("naca0010").local_thickness(xi)
        )
        for xi in np.cosspace(0, 1, 20)
    ],
)

fig, ax = plt.subplots(figsize=(7, 6))
Beta, Alpha = np.meshgrid(
    np.linspace(-90, 90, 500),
    np.linspace(-90, 90, 500)
)
aero = asb.AeroBuildup(
    airplane=asb.Airplane(
        fuselages=[fuselage]
    ),
    op_point=asb.OperatingPoint(
        velocity=10,
        alpha=Alpha,
        beta=Beta,
    )
).run()

from aerosandbox.tools.string_formatting import eng_string

p.contour(
    Beta, Alpha, aero["L"], colorbar_label="Lift $L$ [N]",
    # levels=100,
    linelabels_format=lambda s: eng_string(s, unit="N"),
    cmap=plt.get_cmap("coolwarm")
)
p.equal()
p.show_plot("3D Fuselage Lift", r"$\beta$ [deg]", r"$\alpha$ [deg]")



import aerosandbox as asb
import aerosandbox.numpy as np
import pytest


def test_rocket():
    ### Environment
    opti = asb.Opti()

    ### Time discretization
    N = 500  # Number of discretization points
    time_final = 100  # seconds
    time = np.linspace(0, time_final, N)

    ### Constants
    mass_initial = 500e3  # Initial mass, 500 metric tons
    z_e_final = -100e3  # Final altitude, 100 km
    g = 9.81  # Gravity, m/s^2
    alpha = 1 / (300 * g)  # kg/(N*s), Inverse of specific impulse, basically - don't worry about this

    dyn = asb.DynamicsPointMass1DVertical(
        mass_props=asb.MassProperties(mass=opti.variable(init_guess=mass_initial, n_vars=N)),
        z_e=opti.variable(init_guess=np.linspace(0, z_e_final, N)),  # Altitude (negative due to Earth-axes convention)
        w_e=opti.variable(init_guess=-z_e_final / time_final, n_vars=N),  # Velocity
    )

    dyn.add_gravity_force(g=g)
    thrust = opti.variable(init_guess=g * mass_initial, n_vars=N)
    dyn.add_force(Fz=-thrust)

    dyn.constrain_derivatives(
        opti=opti,
        time=time,
    )

    ### Fuel burn
    opti.constrain_derivative(
        derivative=-alpha * thrust,
        variable=dyn.mass_props.mass,
        with_respect_to=time,
        method="midpoint",
    )

    ### Boundary conditions
    opti.subject_to([
        dyn.z_e[0] == 0,
        dyn.w_e[0] == 0,
        dyn.mass_props.mass[0] == mass_initial,
        dyn.z_e[-1] == z_e_final,
    ])

    ### Path constraints
    opti.subject_to([
        dyn.mass_props.mass >= 0,
        thrust >= 0
    ])

    ### Objective
    opti.minimize(-dyn.mass_props.mass[-1])  # Maximize the final mass == minimize fuel expenditure

    ### Solve
    sol = opti.solve(verbose=False)
    print(f"Solved in {sol.stats()['iter_count']} iterations.")
    dyn.substitute_solution(sol)

    assert dyn.mass_props.mass[-1] == pytest.approx(290049.81034472014, rel=0.05)
    assert np.abs(dyn.w_e).max() == pytest.approx(1448, rel=0.05)


if __name__ == '__main__':
    test_rocket()
    pytest.main()



import aerosandbox.numpy as np
import datetime
from aerosandbox.modeling.interpolation import InterpolatedModel
from pathlib import Path
import os


def wind_speed_conus_summer_99(altitude, latitude):
    """
    Returns the 99th-percentile wind speed magnitude over the continental United States (CONUS) in July-Aug. Aggregate of data from 1972 to 2019.
    Fits at C:\Projects\GitHub\Wind_Analysis
    :param altitude: altitude [m]
    :param latitude: latitude [deg]
    :return: 99th-percentile wind speed over the continental United States in the summertime. [m/s]
    """
    l = (latitude - 37.5) / 11.5
    a = (altitude - 24200) / 24200

    agc = -0.5363486000267786
    agh = 1.9569754777072828
    ags = 0.1458701999734713
    aqc = -1.4645014948089652
    c0 = -0.5169694086686631
    c12 = 0.0849519807021402
    c21 = -0.0252010113059998
    c4a = 0.0225856848053377
    c4c = 1.0281877353734501
    cg = 0.8050736230004489
    cgc = 0.2786691793571486
    cqa = 0.1866078047914259
    cql = 0.0165126852561671
    cqla = -0.1361667658248024
    lgc = 0.6943655538727291
    lgh = 2.0777449841036777
    lgs = 0.9805766577269118
    lqc = 4.0356834595743214

    s = c0 + cql * (l - lqc) ** 2 + cqa * (a - aqc) ** 2 + cqla * a * l + cg * np.exp(
        -(np.fabs(l - lgc) ** lgh / (2 * lgs ** 2) + np.fabs(a - agc) ** agh / (
                2 * ags ** 2) + cgc * a * l)) + c4a * (
                a - c4c) ** 4 + c12 * l * a ** 2 + c21 * l ** 2 * a

    speed = s * 56 + 7
    return speed


### Prep data for global wind speed function
# Import data
root = Path(os.path.abspath(__file__)).parent
altitudes_world = np.load(root / "datasets" / "winds_and_tropopause_global" / "altitudes.npy")
latitudes_world = np.load(root / "datasets" / "winds_and_tropopause_global" / "latitudes.npy")
day_of_year_world_boundaries = np.linspace(0, 365, 13)
day_of_year_world = (day_of_year_world_boundaries[1:] + day_of_year_world_boundaries[:-1]) / 2
winds_95_world = np.load(root / "datasets" / "winds_and_tropopause_global" / "winds_95_vs_altitude_latitude_day.npy")

# Trim the poles
latitudes_world = latitudes_world[1:-1]
winds_95_world = winds_95_world[:, 1:-1, :]

# Flip data appropriately
altitudes_world = np.flip(altitudes_world)
latitudes_world = np.flip(latitudes_world)
### NOTE: winds_95_world has *already* been flipped appropriately

# # Extend altitude range down to the ground # TODO review and redo properly
# altitudes_world_to_extend = [-1000, 0, 5000]
# altitudes_world = np.hstack((
#     altitudes_world_to_extend,
#     altitudes_world
# ))
# winds_95_world = np.concatenate(
#     (
#         np.tile(
#             winds_95_world[0, :, :],
#             (3, 1, 1)
#         ),
#         winds_95_world
#     ),
#     axis=0
# )

# Downsample
latitudes_world = latitudes_world[::5]
winds_95_world = winds_95_world[:, ::5, :]

# Extend boundaries so that cubic spline interpolates around day_of_year appropriately.
extend_bounds = 3
day_of_year_world = np.hstack((
    day_of_year_world[-extend_bounds:] - 365,
    day_of_year_world,
    day_of_year_world[:extend_bounds] + 365
))
winds_95_world = np.dstack((
    winds_95_world[:, :, -extend_bounds:],
    winds_95_world,
    winds_95_world[:, :, :extend_bounds]
))

# Make the model
winds_95_world_model = InterpolatedModel(
    x_data_coordinates={
        "altitude"   : altitudes_world,
        "latitude"   : latitudes_world,
        "day of year": day_of_year_world,
    },
    y_data_structured=winds_95_world,
)


def wind_speed_world_95(
        altitude,
        latitude,
        day_of_year
):
    """
    Gives the 95th-percentile wind speed as a function of altitude, latitude, and day of year.
    Args:
        altitude: Altitude, in meters
        latitude: Latitude, in degrees north
        day_of_year: Day of year (Julian day), in range 0 to 365

    Returns: The 95th-percentile wind speed, in meters per second.

    """

    return winds_95_world_model({
        "altitude"   : altitude,
        "latitude"   : latitude,
        "day of year": day_of_year
    })


### Prep data for tropopause altitude function
# Import data
latitudes_trop = np.linspace(-80, 80, 50)
day_of_year_trop_boundaries = np.linspace(0, 365, 13)
day_of_year_trop = (day_of_year_trop_boundaries[1:] + day_of_year_trop_boundaries[:-1]) / 2
tropopause_altitude_km = np.genfromtxt(
    root / "datasets" / "winds_and_tropopause_global" / "strat-height-monthly.csv",
    delimiter=","
)

# Extend boundaries
extend_bounds = 3
day_of_year_trop = np.hstack((
    day_of_year_trop[-extend_bounds:] - 365,
    day_of_year_trop,
    day_of_year_trop[:extend_bounds] + 365
))
tropopause_altitude_km = np.hstack((
    tropopause_altitude_km[:, -extend_bounds:],
    tropopause_altitude_km,
    tropopause_altitude_km[:, :extend_bounds]
))

# Make the model
tropopause_altitude_model = InterpolatedModel(
    x_data_coordinates={
        "latitude"   : latitudes_trop,
        "day of year": day_of_year_trop
    },
    y_data_structured=tropopause_altitude_km * 1e3
)


def tropopause_altitude(
        latitude,
        day_of_year
):
    """
    Gives the altitude of the tropopause (as determined by the altitude where lapse rate >= 2 C/km) as a function of
    latitude and day of year.

    Args:
        altitude: Altitude, in meters
        day_of_year: Day of year (Julian day), in range 0 to 365

    Returns: The tropopause altitude, in meters.

    """
    return tropopause_altitude_model({
        "latitude"   : latitude,
        "day of year": day_of_year
    })


if __name__ == '__main__':

    from aerosandbox.tools.pretty_plots import plt, sns, mpl, show_plot


    def plot_winds_at_altitude(altitude=18000):
        fig, ax = plt.subplots()

        day_of_years = np.linspace(0, 365, 150)
        latitudes = np.linspace(-80, 80, 120)
        Day_of_years, Latitudes = np.meshgrid(day_of_years, latitudes)

        winds = wind_speed_world_95(
            altitude=altitude * np.ones_like(Latitudes.flatten()),
            latitude=Latitudes.flatten(),
            day_of_year=Day_of_years.flatten(),
        ).reshape(Latitudes.shape)

        args = [
            day_of_years,
            latitudes,
            winds
        ]

        levels = np.arange(0, 80.1, 5)
        CS = plt.contour(*args, levels=levels, linewidths=0.5, colors="k", alpha=0.7)
        CF = plt.contourf(*args, levels=levels, cmap='viridis_r', alpha=0.7, extend="max")
        cbar = plt.colorbar(label="Wind Speed [m/s]", extendrect=True)
        ax.clabel(CS, inline=1, fontsize=9, fmt="%.0f m/s")

        plt.xticks(
            np.linspace(0, 365, 13)[:-1],
            (
                "Jan. 1",
                "Feb. 1",
                "Mar. 1",
                "Apr. 1",
                "May 1",
                "June 1",
                "July 1",
                "Aug. 1",
                "Sep. 1",
                "Oct. 1",
                "Nov. 1",
                "Dec. 1"
            ),
            rotation=40
        )

        lat_label_vals = np.arange(-80, 80.1, 20)
        lat_labels = []
        for lat in lat_label_vals:
            if lat >= 0:
                lat_labels.append(f"{lat:.0f}N")
            else:
                lat_labels.append(f"{-lat:.0f}S")
        plt.yticks(
            lat_label_vals,
            lat_labels
        )

        show_plot(
            f"95th-Percentile Wind Speeds at {altitude / 1e3:.0f} km Altitude",
            xlabel="Day of Year",
            ylabel="Latitude",
        )


    def plot_winds_at_day(day_of_year=0):
        fig, ax = plt.subplots()

        altitudes = np.linspace(0, 30000, 150)
        latitudes = np.linspace(-80, 80, 120)
        Altitudes, Latitudes = np.meshgrid(altitudes, latitudes)

        winds = wind_speed_world_95(
            altitude=Altitudes.flatten(),
            latitude=Latitudes.flatten(),
            day_of_year=day_of_year * np.ones_like(Altitudes.flatten()),
        ).reshape(Altitudes.shape)

        args = [
            altitudes / 1e3,
            latitudes,
            winds
        ]

        levels = np.arange(0, 80.1, 5)
        CS = plt.contour(*args, levels=levels, linewidths=0.5, colors="k", alpha=0.7)
        CF = plt.contourf(*args, levels=levels, cmap='viridis_r', alpha=0.7, extend="max")
        cbar = plt.colorbar(label="Wind Speed [m/s]", extendrect=True)
        ax.clabel(CS, inline=1, fontsize=9, fmt="%.0f m/s")

        lat_label_vals = np.arange(-80, 80.1, 20)
        lat_labels = []
        for lat in lat_label_vals:
            if lat >= 0:
                lat_labels.append(f"{lat:.0f}N")
            else:
                lat_labels.append(f"{-lat:.0f}S")
        plt.yticks(
            lat_label_vals,
            lat_labels
        )

        show_plot(
            f"95th-Percentile Wind Speeds at Day {day_of_year:.0f}",
            xlabel="Altitude [km]",
            ylabel="Latitude",
        )


    def plot_tropopause_altitude():
        fig, ax = plt.subplots()

        day_of_years = np.linspace(0, 365, 250)
        latitudes = np.linspace(-80, 80, 200)
        Day_of_years, Latitudes = np.meshgrid(day_of_years, latitudes)

        trop_alt = tropopause_altitude(
            Latitudes.flatten(),
            Day_of_years.flatten()
        ).reshape(Latitudes.shape)

        args = [
            day_of_years,
            latitudes,
            trop_alt / 1e3
        ]

        levels = np.arange(10, 20.1, 1)
        CS = plt.contour(*args, levels=levels, linewidths=0.5, colors="k", alpha=0.7)
        CF = plt.contourf(*args, levels=levels, cmap='viridis_r', alpha=0.7, extend="both")
        cbar = plt.colorbar(label="Tropopause Altitude [km]", extendrect=True)
        ax.clabel(CS, inline=1, fontsize=9, fmt="%.0f km")

        plt.xticks(
            np.linspace(0, 365, 13)[:-1],
            (
                "Jan. 1",
                "Feb. 1",
                "Mar. 1",
                "Apr. 1",
                "May 1",
                "June 1",
                "July 1",
                "Aug. 1",
                "Sep. 1",
                "Oct. 1",
                "Nov. 1",
                "Dec. 1"
            ),
            rotation=40
        )

        lat_label_vals = np.arange(-80, 80.1, 20)
        lat_labels = []
        for lat in lat_label_vals:
            if lat >= 0:
                lat_labels.append(f"{lat:.0f}N")
            else:
                lat_labels.append(f"{-lat:.0f}S")
        plt.yticks(
            lat_label_vals,
            lat_labels
        )

        show_plot(
            f"Tropopause Altitude by Season and Latitude",
            xlabel="Day of Year",
            ylabel="Latitude",
        )


    def plot_winds_at_tropopause_altitude():
        fig, ax = plt.subplots()

        day_of_years = np.linspace(0, 365, 150)
        latitudes = np.linspace(-80, 80, 120)
        Day_of_years, Latitudes = np.meshgrid(day_of_years, latitudes)

        winds = wind_speed_world_95(
            altitude=tropopause_altitude(Latitudes.flatten(), Day_of_years.flatten()),
            latitude=Latitudes.flatten(),
            day_of_year=Day_of_years.flatten(),
        ).reshape(Latitudes.shape)

        args = [
            day_of_years,
            latitudes,
            winds
        ]

        levels = np.arange(0, 80.1, 5)
        CS = plt.contour(*args, levels=levels, linewidths=0.5, colors="k", alpha=0.7)
        CF = plt.contourf(*args, levels=levels, cmap='viridis_r', alpha=0.7, extend="max")
        cbar = plt.colorbar(label="Wind Speed [m/s]", extendrect=True)
        ax.clabel(CS, inline=1, fontsize=9, fmt="%.0f m/s")

        plt.xticks(
            np.linspace(0, 365, 13)[:-1],
            (
                "Jan. 1",
                "Feb. 1",
                "Mar. 1",
                "Apr. 1",
                "May 1",
                "June 1",
                "July 1",
                "Aug. 1",
                "Sep. 1",
                "Oct. 1",
                "Nov. 1",
                "Dec. 1"
            ),
            rotation=40
        )

        lat_label_vals = np.arange(-80, 80.1, 20)
        lat_labels = []
        for lat in lat_label_vals:
            if lat >= 0:
                lat_labels.append(f"{lat:.0f}N")
            else:
                lat_labels.append(f"{-lat:.0f}S")
        plt.yticks(
            lat_label_vals,
            lat_labels
        )

        show_plot(
            f"95th-Percentile Wind Speeds at Tropopause Altitude",
            xlabel="Day of Year",
            ylabel="Latitude",
        )


    # plot_winds_at_altitude(altitude=18000)
    # plot_winds_at_day(day_of_year=0)
    # plot_tropopause_altitude()
    plot_winds_at_tropopause_altitude()



import aerosandbox as asb
import aerosandbox.numpy as np
from aerosandbox.tools import units as u
from typing import Union
import pytest


def a() -> asb.Airplane:

    fuselage_cabin_diameter = 20.4 * u.foot
    fuselage_cabin_radius = fuselage_cabin_diameter / 2
    fuselage_cabin_xsec_area = np.pi * fuselage_cabin_radius ** 2

    fuselage_cabin_length = 123.2 * u.foot
    fwd_fuel_tank_length = 6
    aft_fuel_tank_length = fwd_fuel_tank_length

    # Compute x-locations of various fuselage stations
    nose_fineness_ratio = 1.67
    tail_fineness_ratio = 2.62

    x_nose = 0
    x_nose_to_fwd_tank = x_nose + nose_fineness_ratio * fuselage_cabin_diameter
    x_fwd_tank_to_cabin = x_nose_to_fwd_tank + fwd_fuel_tank_length
    x_cabin_to_aft_tank = x_fwd_tank_to_cabin + fuselage_cabin_length
    x_aft_tank_to_tail = x_cabin_to_aft_tank + aft_fuel_tank_length
    x_tail = x_aft_tank_to_tail + tail_fineness_ratio * fuselage_cabin_diameter

    # Build up the actual fuselage nodes
    x_fuse_sections = []
    z_fuse_sections = []
    r_fuse_sections = []

    def linear_map(
            f_in: Union[float, np.ndarray],
            min_in: Union[float, np.ndarray],
            max_in: Union[float, np.ndarray],
            min_out: Union[float, np.ndarray],
            max_out: Union[float, np.ndarray],
    ) -> Union[float, np.ndarray]:
        """
        Linearly maps an input `f_in` from range (`min_in`, `max_in`) to (`min_out`, `max_out`).

        Args:
            f_in: Input value
            min_in:
            max_in:
            min_out:
            max_out:

        Returns:
            f_out: Output value

        """
        # if min_in == 0 and max_in == 1:
        #     f_nondim = f_in
        # else:
        f_nondim = (f_in - min_in) / (max_in - min_in)

        # if max_out == 0 and min_out == 1:
        #     f_out = f_nondim
        # else:
        f_out = f_nondim * (max_out - min_out) + min_out

        return f_out

    # Nose
    x_sect_nondim = np.sinspace(0, 1, 10)
    z_sect_nondim = -0.3 * (1 - x_sect_nondim) ** 2
    r_sect_nondim = (1 - (1 - x_sect_nondim) ** 2) ** 0.5

    x_fuse_sections.append(
        linear_map(
            f_in=x_sect_nondim,
            min_in=0, max_in=1,
            min_out=x_nose, max_out=x_nose_to_fwd_tank
        )
    )
    z_fuse_sections.append(
        linear_map(
            f_in=z_sect_nondim,
            min_in=0, max_in=1,
            min_out=0, max_out=fuselage_cabin_radius
        )
    )
    r_fuse_sections.append(
        linear_map(
            f_in=r_sect_nondim,
            min_in=0, max_in=1,
            min_out=0, max_out=fuselage_cabin_radius
        )
    )

    # Fwd tank
    x_sect_nondim = np.linspace(0, 1, 2)
    z_sect_nondim = np.zeros_like(x_sect_nondim)
    r_sect_nondim = np.ones_like(x_sect_nondim)

    x_fuse_sections.append(
        linear_map(
            f_in=x_sect_nondim,
            min_in=0, max_in=1,
            min_out=x_nose_to_fwd_tank, max_out=x_fwd_tank_to_cabin
        )
    )
    z_fuse_sections.append(
        linear_map(
            f_in=z_sect_nondim,
            min_in=0, max_in=1,
            min_out=0, max_out=fuselage_cabin_radius
        )
    )
    r_fuse_sections.append(
        linear_map(
            f_in=r_sect_nondim,
            min_in=0, max_in=1,
            min_out=0, max_out=fuselage_cabin_radius
        )
    )

    # Cabin
    x_sect_nondim = np.linspace(0, 1, 2)
    z_sect_nondim = np.zeros_like(x_sect_nondim)
    r_sect_nondim = np.ones_like(x_sect_nondim)

    x_fuse_sections.append(
        linear_map(
            f_in=x_sect_nondim,
            min_in=0, max_in=1,
            min_out=x_fwd_tank_to_cabin, max_out=x_cabin_to_aft_tank
        )
    )
    z_fuse_sections.append(
        linear_map(
            f_in=z_sect_nondim,
            min_in=0, max_in=1,
            min_out=0, max_out=fuselage_cabin_radius
        )
    )
    r_fuse_sections.append(
        linear_map(
            f_in=r_sect_nondim,
            min_in=0, max_in=1,
            min_out=0, max_out=fuselage_cabin_radius
        )
    )

    # Aft Tank
    x_sect_nondim = np.linspace(0, 1, 2)
    z_sect_nondim = np.zeros_like(x_sect_nondim)
    r_sect_nondim = np.ones_like(x_sect_nondim)

    x_fuse_sections.append(
        linear_map(
            f_in=x_sect_nondim,
            min_in=0, max_in=1,
            min_out=x_cabin_to_aft_tank, max_out=x_aft_tank_to_tail
        )
    )
    z_fuse_sections.append(
        linear_map(
            f_in=z_sect_nondim,
            min_in=0, max_in=1,
            min_out=0, max_out=fuselage_cabin_radius
        )
    )
    r_fuse_sections.append(
        linear_map(
            f_in=r_sect_nondim,
            min_in=0, max_in=1,
            min_out=0, max_out=fuselage_cabin_radius
        )
    )

    # Tail
    x_sect_nondim = np.linspace(0, 1, 10)
    z_sect_nondim = 1 * x_sect_nondim ** 1.5
    r_sect_nondim = 1 - x_sect_nondim ** 1.5

    x_fuse_sections.append(
        linear_map(
            f_in=x_sect_nondim,
            min_in=0, max_in=1,
            min_out=x_aft_tank_to_tail, max_out=x_tail
        )
    )
    z_fuse_sections.append(
        linear_map(
            f_in=z_sect_nondim,
            min_in=0, max_in=1,
            min_out=0, max_out=fuselage_cabin_radius
        )
    )
    r_fuse_sections.append(
        linear_map(
            f_in=r_sect_nondim,
            min_in=0, max_in=1,
            min_out=0, max_out=fuselage_cabin_radius
        )
    )

    # Compile Fuselage
    x_fuse_sections = np.concatenate([
        x_fuse_section[:-1] if i != len(x_fuse_sections) - 1 else x_fuse_section
        for i, x_fuse_section in enumerate(x_fuse_sections)
    ])
    z_fuse_sections = np.concatenate([
        z_fuse_section[:-1] if i != len(z_fuse_sections) - 1 else z_fuse_section
        for i, z_fuse_section in enumerate(z_fuse_sections)
    ])
    r_fuse_sections = np.concatenate([
        r_fuse_section[:-1] if i != len(r_fuse_sections) - 1 else r_fuse_section
        for i, r_fuse_section in enumerate(r_fuse_sections)
    ])

    fuse = asb.Fuselage(
        name="Fuselage",
        xsecs=[
            asb.FuselageXSec(
                xyz_c=[
                    x_fuse_sections[i],
                    0,
                    z_fuse_sections[i]
                ],
                radius=r_fuse_sections[i]
            )
            for i in range(np.length(x_fuse_sections))
        ],
        analysis_specific_options={
            asb.AeroBuildup: dict(
                nose_fineness_ratio=nose_fineness_ratio
            )
        }
    )

    ### Wing
    wing_airfoil = asb.Airfoil("b737c").repanel(100)
    wing_airfoil.generate_polars(
        cache_filename="cache/b737c.json",
        include_compressibility_effects=True,
    )

    wing_span = 214 * u.foot
    wing_half_span = wing_span / 2

    wing_root_chord = 51.5 * u.foot

    wing_LE_sweep_deg = 34

    wing_yehudi_span_fraction = 0.25
    wing_dihedral = 6

    # Compute the y locations
    wing_yehudi_y = wing_yehudi_span_fraction * wing_half_span
    wing_tip_y = wing_half_span

    # Compute the x locations
    wing_yehudi_x = wing_yehudi_y * np.tand(wing_LE_sweep_deg)
    wing_tip_x = wing_tip_y * np.tand(wing_LE_sweep_deg)

    # Compute the chords
    wing_yehudi_chord = wing_root_chord - wing_yehudi_x
    wing_tip_chord = 0.14 * wing_root_chord

    # Make the sections
    wing_root = asb.WingXSec(
        xyz_le=[0, 0, 0],
        chord=wing_root_chord,
        airfoil=wing_airfoil,
    )
    wing_yehudi = asb.WingXSec(
        xyz_le=[
            wing_yehudi_x,
            wing_yehudi_y,
            wing_yehudi_y * np.tand(wing_dihedral)
        ],
        chord=wing_yehudi_chord,
        airfoil=wing_airfoil,
    )
    wing_tip = asb.WingXSec(
        xyz_le=[
            wing_tip_x,
            wing_tip_y,
            wing_tip_y * np.tand(wing_dihedral)
        ],
        chord=wing_tip_chord,
        airfoil=wing_airfoil
    )

    # Assemble the wing
    wing_x_le = 0.5 * x_fwd_tank_to_cabin + 0.5 * x_cabin_to_aft_tank - 0.5 * wing_root_chord

    wing_z_le = -0.5 * fuselage_cabin_radius

    wing = asb.Wing(
        name="Main Wing",
        symmetric=True,
        xsecs=[
            wing_root,
            wing_yehudi,
            wing_tip
        ]
    ).translate([
        wing_x_le,
        0,
        wing_z_le
    ]).subdivide_sections(3)

    ### Horizontal Stabilizer
    hstab_airfoil = asb.Airfoil("naca0012")
    hstab_airfoil.generate_polars(
        cache_filename="cache/naca0012.json",
        include_compressibility_effects=True
    )

    hstab_span = 70.8 * u.foot
    hstab_half_span = hstab_span / 2

    hstab_root_chord = 23 * u.foot

    hstab_LE_sweep_deg = 39

    elevator = asb.ControlSurface(
        name="All-moving Elevator",
        deflection=0
    )

    hstab_root = asb.WingXSec(
        xyz_le=[0, 0, 0],
        chord=hstab_root_chord,
        airfoil=hstab_airfoil,
        control_surfaces=[
            elevator
        ]
    )
    hstab_tip = asb.WingXSec(
        xyz_le=[
            hstab_half_span * np.tand(hstab_LE_sweep_deg),
            hstab_half_span,
            0
        ],
        chord=0.35 * hstab_root_chord,
        airfoil=hstab_airfoil
    )

    # Assemble the hstab
    hstab_x_le = x_tail - 2 * hstab_root_chord
    hstab_z_le = 0.5 * fuselage_cabin_radius

    hstab = asb.Wing(
        name="Horizontal Stabilizer",
        symmetric=True,
        xsecs=[
            hstab_root,
            hstab_tip
        ]
    ).translate([
        hstab_x_le,
        0,
        hstab_z_le
    ]).subdivide_sections(3)

    ### Vertical Stabilizer
    vstab_airfoil = asb.Airfoil("naca0008")
    vstab_airfoil.generate_polars(
        cache_filename="cache/naca0008.json",
        include_compressibility_effects=True
    )

    vstab_span = 29.6 * u.foot

    vstab_root_chord = 22 * u.foot

    vstab_LE_sweep_deg = 45

    vstab_root = asb.WingXSec(
        xyz_le=[0, 0, 0],
        chord=vstab_root_chord,
        airfoil=vstab_airfoil
    )
    vstab_tip = asb.WingXSec(
        xyz_le=[
            vstab_span * np.tand(vstab_LE_sweep_deg),
            0,
            vstab_span,
        ],
        chord=0.35 * vstab_root_chord,
        airfoil=vstab_airfoil
    )

    # Assemble the vstab
    vstab_x_le = x_tail - 2 * vstab_root_chord
    vstab_z_le = 1 * fuselage_cabin_radius

    vstab = asb.Wing(
        name="Vertical Stabilizer",
        xsecs=[
            vstab_root,
            vstab_tip
        ]
    ).translate([
        vstab_x_le,
        0,
        vstab_z_le
    ]).subdivide_sections(3)

    ### Airplane
    airplane = asb.Airplane(
        name="Airplane",
        xyz_ref=[],
        wings=[
            wing,
            hstab,
            vstab
        ],
        fuselages=[
            fuse
        ],
    )

    return airplane


if __name__ == '__main__':
    # import matplotlib
    # matplotlib.use("TkAgg")
    import matplotlib.pyplot as plt
    import aerosandbox.tools.pretty_plots as p

    airplane = a()

    airplane.draw_three_view()

    # airplane.draw(backend="matplotlib", show=False)
    #
    # import matplotlib.pyplot as plt
    # import aerosandbox.tools.pretty_plots as p
    #
    # ax = plt.gca()
    #
    # t = np.linspace(0, 1, 100)
    #
    # x = airplane.fuselages[0].length() * t
    # y = 15 * np.sin(4 * 2 * np.pi * t)
    # z = 15 *np.cos(4 * 2 * np.pi * t)
    #
    # ax.plot(
    #     x, y, z
    # )
    #
    # p.show_plot()
